<!DOCTYPE html >
<html lang="en">
<head>
    <meta charset="UTF-8">
    <script type="text/javascript" src="https://cdnjs.cloudflare.com/ajax/libs/jquery/3.1.1/jquery.min.js"></script>
    <script type="text/javascript" src="https://cdnjs.cloudflare.com/ajax/libs/d3/4.5.0/d3.min.js"></script>
    <script type="text/javascript">
        {% include "pybel_vis.js" %}
    </script>
    <script type="text/javascript">
        graph = {{ graph|safe }};
        colorMap = {{ color_map|safe }};

        $(document).ready(function () {
            w = $("#graph-chart").width();
            g = document.getElementsByTagName('body')[0];
            h = w.innerHeight || document.documentElement.clientHeight || g.clientHeight;
            init_d3_force(d3, graph, "#graph-chart", w, h, colorMap);
        });
    </script>
</head>
<body>
<div id="graph-chart"></div>
</body>
</html>
