# -*- coding: utf-8 -*-

"""Resources that don't change over time for PyBEL."""

SPECIES_PATTERN = r"^\d+$"
CONFIDENCE_URL = "https://arty.scai.fraunhofer.de/artifactory/bel/annotation/confidence/confidence-1.0.0.belanno"
